/* -------------------------------------------------------------------------- */
/* Project: I Curso de Computación Científica en Clusters                     */
/* Author:  Juan Fernández Peinador                                           */
/* Date:    Marzo de 2010                                                     */
/* -------------------------------------------------------------------------- */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

// includes, project
#include <cutil_inline.h>

// includes, kernels
#include "cuda_matrixAdd_kernel.cu"

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    float *matrix1_h, *matrix2_h, *matrixR_h; // host data
    float *matrix1_d, *matrix2_d, *matrixR_d; // device data
    long int nPos;
    size_t nBytes;

    // default values
    int n = 1;
    int bsx = 1;

    // General initialization call to pick the best CUDA Device
    // If the command-line has a device number specified, use it
    // Otherwise pick the device with highest Gflops/s
    cutilChooseCudaDevice(argc, argv);

    // process command line arguments
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "n", &n);
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "bsx", &bsx);

    nPos = n * n;
    nBytes = nPos * sizeof(float);

    // allocate host memory
    matrix1_h = (float *) malloc(nBytes);
    for(int i = 0; i < nPos; i++)
        matrix1_h[i] = (float) 1.0;

    matrix2_h = (float *) malloc(nBytes);
    for(int i = 0; i < nPos; i++)
        matrix2_h[i] = (float) 1.0;
    
    matrixR_h = (float *) malloc(nBytes);
    bzero(matrixR_h, nBytes);

    // allocate device memory
    cutilSafeCall(cudaMalloc((void **) &matrix1_d, nBytes));
    cutilSafeCall(cudaMalloc((void **) &matrix2_d, nBytes));
    cutilSafeCall(cudaMalloc((void **) &matrixR_d, nBytes));

    // copy data from host memory to device memory
    cutilSafeCall(cudaMemcpy(matrix1_d, matrix1_h, nBytes, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(matrix2_d, matrix2_h, nBytes, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemset(matrixR_d, 0, nBytes));

    // create and start timer
    unsigned int timer = 0;
    cutilCheckError(cutCreateTimer(&timer));
    cutilCheckError(cutStartTimer(timer));

    // setup execution parameters
    dim3 grid( (nPos%bsx) ? (nPos/bsx)+1 : (nPos/bsx) );
    dim3 block(bsx);

    // execute the kernel
    printf("Running configuration: grid of %d x %d blocks of %d x %d threads (%d threads)\n", 
           grid.x, grid.y, block.x, block.y, grid.x * grid.y * block.x * block.y);

    //matrixAdd<<<grid, block>>>(matrix1_d, matrix2_d, matrixR_d, n, m);
    matrixAdd<<<grid, block>>>(matrix1_d, matrix2_d, matrixR_d, n);

    // wait for thread completion
    cudaThreadSynchronize();

    // check if kernel execution generated any error
    cutilCheckMsg("Kernel execution failed");

    cutilSafeCall(cudaMemcpy(matrixR_h, matrixR_d, nBytes, cudaMemcpyDeviceToHost));

    // stop and destroy timer
    cutilCheckError(cutStopTimer(timer));
    printf("Processing time: %f (ms)", cutGetTimerValue(timer));
    cutilCheckError(cutDeleteTimer(timer));

    // check results
    for(int i = 0; i < nPos; i++)
        assert(matrixR_h[i] == (float) 2.0);

    // free memory
    free(matrix1_h);
    free(matrix2_h);
    free(matrixR_h);
    cutilSafeCall(cudaFree((void *) matrix1_d));
    cutilSafeCall(cudaFree((void *) matrix2_d));
    cutilSafeCall(cudaFree((void *) matrixR_d));

    printf("\nTest PASSED\n");
    cudaThreadExit();
}

