/* -------------------------------------------------------------------------- */
/* Project: I Curso de Computación Científica en Clusters                     */
/* Author:  Juan Fernández Peinador                                           */
/* Date:    Marzo de 2010                                                     */
/* -------------------------------------------------------------------------- */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

// includes, project
#include <cutil_inline.h>

#define SHM_SIZE (16 * 1024)
#define CM_SIZE  (8)

// includes, kernels
#include "cuda_template_kernel.cu"

const int constante_h[CM_SIZE] = {1, 2, 3, 4, 5, 6, 7, 8};

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    int *gid_h; // host data
    int *gid_d; // device data
    size_t shared_mem_size;
    long int nPos;
    size_t nBytes;

    // default values
    int tam_grid_x  = 1;
    int tam_grid_y  = 1;
    int tam_block_x = 1;
    int tam_block_y = 1;
    //int tam_block_z = 1;

    // General initialization call to pick the best CUDA Device
    // If the command-line has a device number specified, use it
    // Otherwise pick the device with highest Gflops/s
    cutilChooseCudaDevice(argc, argv);

    // process command line arguments
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "gsx", &tam_grid_x);
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "gsy", &tam_grid_y);
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "bsx", &tam_block_x);
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "bsy", &tam_block_y);

    nPos = tam_grid_x * tam_grid_y * tam_block_x * tam_block_y;
    nBytes = nPos * sizeof(int);

    // allocate host memory
    gid_h = (int *) malloc(nBytes);
    bzero(gid_h, nBytes);

    // allocate device memory
    cutilSafeCall(cudaMalloc((void **) &gid_d, nBytes));
   
    // copy data from host memory to device memory
    cutilSafeCall(cudaMemcpy(gid_d, gid_h, nBytes, cudaMemcpyHostToDevice));
    // cutilSafeCall(cudaMemset((void *) gid_d, 0, nBytes));

    // initialize constant memory
    cutilSafeCall(cudaMemcpyToSymbol("constante_d", constante_h, CM_SIZE*sizeof(int), 0, cudaMemcpyHostToDevice));

    // create and start timer
    unsigned int timer = 0;
    cutilCheckError(cutCreateTimer(&timer));
    cutilCheckError(cutStartTimer(timer));

    // setup execution parameters
    dim3 grid(tam_grid_x, tam_grid_y);
    dim3 block(tam_block_x, tam_block_y);

    // execute the kernel
    shared_mem_size = block.x * block.y * sizeof(int);
    assert(shared_mem_size <= SHM_SIZE);
    printf("Running configuration: \t %d threads\n\t\t\t grid of %d x %d\n"
           "\t\t\t blocks of %d x %d threads (%d threads with %d bytes of shared memory per block)\n", 
           nPos,
           tam_grid_x, tam_grid_y, tam_block_x, tam_block_y,
           tam_block_x * tam_block_y, shared_mem_size);
    foo<<<grid, block, shared_mem_size>>>(gid_d);

    // wait for thread completion
    cudaThreadSynchronize();

    // check if kernel execution generated any error
    cutilCheckMsg("Kernel execution failed");

    // get results back from device memory
    cutilSafeCall(cudaMemcpy(gid_h, gid_d, nBytes, cudaMemcpyDeviceToHost));

    // stop and destroy timer
    cutilCheckError(cutStopTimer(timer));
    printf("Processing time: %f (ms)", cutGetTimerValue(timer));
    cutilCheckError(cutDeleteTimer(timer));

    // check results
    for(int i = 0; i < nPos; i++)
        assert(gid_h[i] == (i+constante_h[i%CM_SIZE]));

    // free memory
    free(gid_h);
    cutilSafeCall(cudaFree((void *) gid_d));

    printf("\nTest PASSED\n");
    cudaThreadExit();
}

