/* -------------------------------------------------------------------------- */
/* Project: I Curso de Computación Científica en Clusters                     */
/* Author:  Juan Fernández Peinador                                           */
/* Date:    Marzo de 2010                                                     */
/* -------------------------------------------------------------------------- */

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

// includes, project
#include <cutil_inline.h>

// includes, kernels
#include "cuda_vectorReduce_kernel.cu"

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main(int argc, char **argv)
{
    float *vector_h, *reduce_h; // host data
    float *vector_d, *reduce_d; // device data
    size_t nBytes;

    // default values
    int n = 1;
    int bsx = 1;

    // General initialization call to pick the best CUDA Device
    // If the command-line has a device number specified, use it
    // Otherwise pick the device with highest Gflops/s
    cutilChooseCudaDevice(argc, argv);

    // process command line arguments
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "n", &n);
    cutGetCmdLineArgumenti(argc, (const char **) argv, (const char *) "bsx", &bsx);

    nBytes = n * sizeof(float);

    // setup execution parameters
    dim3 grid( (n%bsx) ? (n/bsx)+1 : (n/bsx) );
    dim3 block(bsx);

    // allocate host memory
    vector_h = (float *) malloc(nBytes);
    for(int i = 0; i < n; i++)
        vector_h[i] = (float) 1.0;
    reduce_h = (float *) malloc(grid.x * sizeof(float));
    bzero(reduce_h, 1 * sizeof(float));
    
    // allocate device memory
    cutilSafeCall(cudaMalloc((void **) &vector_d, nBytes));
    cutilSafeCall(cudaMalloc((void **) &reduce_d, grid.x * sizeof(float)));

    // copy data from host memory to device memory
    cutilSafeCall(cudaMemcpy(vector_d, vector_h, nBytes, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemset(reduce_d, 0, grid.x * sizeof(float)));

    // create and start timer
    unsigned int timer = 0;
    cutilCheckError(cutCreateTimer(&timer));
    cutilCheckError(cutStartTimer(timer));

    // execute the kernel
    printf("Running configuration: grid of %d blocks of %d threads (%d threads)\n", 
           grid.x, block.x, grid.x * block.x );

    vectorReduce<<<grid, block, block.x * sizeof(float)>>>(vector_d, reduce_d, n);
    
    // wait for thread completion
    cudaThreadSynchronize();

    // check if kernel execution generated any error
    cutilCheckMsg("Kernel execution failed");

    cutilSafeCall(cudaMemcpy(reduce_h, reduce_d, grid.x * sizeof(float), cudaMemcpyDeviceToHost));

    // stop and destroy timer
    cutilCheckError(cutStopTimer(timer));
    printf("Processing time: %f (ms)", cutGetTimerValue(timer));
    printf("Measured bandwidth=%.1fGB/s\n", 1E-9*n*sizeof(int)*1000/cutGetTimerValue(timer));
    cutilCheckError(cutDeleteTimer(timer));

    // compute final stage
    for(int i = 1; i < grid.x; i++)
        reduce_h[0] += reduce_h[i];

    // check result
    assert(reduce_h[0] == (float) n);

    // free memory
    free(vector_h);
    free(reduce_h);
    cutilSafeCall(cudaFree((void *) vector_d));
    cutilSafeCall(cudaFree((void *) reduce_d));

    printf("\nTest PASSED\n");
    cudaThreadExit();
}
